import React, { PureComponent } from 'react';
import PropTypes from 'prop-types';
import pick from 'lodash.pick';
import isFunction from 'lodash.isfunction';
import { Portal } from 'react-overlays';
import {
  stringify, generateItemsFromOptions,
  isSearchable, isMinLengthMatched, getResults
} from './utils';
import Menu from './Menu';
import arrowDown from './img/arrow-down.svg';
import {
  DEFAULT_LABELKEY, SELECT_HEIGHT, SELECT_WIDTH, MAXIMUM_DROPDOWN_HEIGHT, MINIMUM_DROPDOWN_WIDTH,
  UP, DOWN, ESC, RETURN, TAB,
  OPTIONS_PROPS,
  ACTUAL_VALUE
} from './constants';

const isUpdateRequired = (props, state) => {
  const options = generateItemsFromOptions(props.options, props.labelKey);
  const selected = generateItemsFromOptions(props.selected, props.labelKey);
  return (
    (stringify(options) !== stringify(state.options)) ||
    (stringify(selected) !== stringify(state.selected))
  );
};

const getValue = (state, props) => {
  let { selected = [], value = '' } = state;
  const { labelKey } = props;
  selected = [...selected].shift();
  if (selected && typeof selected[labelKey] !== 'undefined') {
    return selected[labelKey];
  }
  return value;
}

const getActiveIndex = (options = [], selected = false, props) => {
  const { labelKey, defaultFirstItemSelected, async } = props;
  if (!selected) {
    return defaultFirstItemSelected && options.length ? 0 : -1;
  }
  if (async) {
    return defaultFirstItemSelected ? 0 : -1;
  }
  return options.findIndex(option => option[labelKey] === selected[labelKey]);
};

const initialStateFromProps = props => {
  const options = [...generateItemsFromOptions(props.options, props.labelKey)];
  const selected = [...generateItemsFromOptions(props.selected, props.labelKey)];
  return {
    options,
    results: [...options],
    selected,
    value: getValue({ selected }, props),
    activeIndex: getActiveIndex(options, selected.length && selected[0], pick(props, ['labelKey', 'defaultFirstItemSelected', 'async'])),
  }
};

class Select extends PureComponent {
  static propTypes = {
    options: PropTypes.array.isRequired,
    selected: PropTypes.array.isRequired,
    labelKey: PropTypes.oneOfType([PropTypes.string, PropTypes.func]),
    searchable: PropTypes.bool,
    onChange: PropTypes.func.isRequired,
    placeHolder: PropTypes.string,
    closeOnOutsideClick: PropTypes.bool,
    multiple: PropTypes.bool,
    maxDropdownHeight: PropTypes.oneOfType([PropTypes.string, PropTypes.number]),
    minDropdownWidth: PropTypes.oneOfType([PropTypes.string, PropTypes.number]),
    width: PropTypes.oneOfType([PropTypes.string, PropTypes.number]),
    height: PropTypes.oneOfType([PropTypes.string, PropTypes.number]),
    minLength: PropTypes.number,
    filterBy: PropTypes.oneOfType([PropTypes.array, PropTypes.func]),
    defaultFirstItemSelected: PropTypes.bool,
    onInputChange: PropTypes.func,
    loading: PropTypes.bool,
    inputProps: PropTypes.object,
    onFocus: PropTypes.func,
    onBlur: PropTypes.func,
  };

  static defaultProps = {
    options: [],
    selected: [],
    labelKey: DEFAULT_LABELKEY,
    searchable: false,
    onChange: () => { },
    placeHolder: 'Select...',
    maxDropdownHeight: MAXIMUM_DROPDOWN_HEIGHT,
    minDropdownWidth: MINIMUM_DROPDOWN_WIDTH,
    width: SELECT_WIDTH,
    height: SELECT_HEIGHT,
    closeOnOutsideClick: true,
    multiple: false,
    minLength: 0,
    filterBy: [],
    defaultFirstItemSelected: false,
    onInputChange: () => { },
    loading: false,
    inputProps: {},
    onFocus: () => { },
    onBlur: () => { },
  };

  state = {
    ...initialStateFromProps(this.props),
    open: false,
    overflow: 'hidden',
    hasFocus: false,
    menuPosition: {
      display: 'none'
    },
    select: null,
    autoScroll: false,
    isDirty: false,
  };

  _input = React.createRef();
  clientWidth = window.innerWidth;
  clientHeight = window.innerHeight;

  componentDidUpdate(props, state) {
    if (state.value !== this.state.value) {
      const results = getResults(this.state.options, this.props, this.state);
      this.setState({
        results
      });
    }
  }

  static getDerivedStateFromProps(props, state) {
    if (
      isUpdateRequired(props, state)
    ) {
      const options = [...generateItemsFromOptions(props.options, props.labelKey)];
      const selected = [...generateItemsFromOptions(props.selected, props.labelKey)];
      const results = getResults(options, props, state);
      const activeIndex = getActiveIndex(results, selected.length && selected[0], pick(props, ['labelKey', 'defaultFirstItemSelected', 'async']));
      return {
        options,
        selected,
        results,
        value: getValue({ selected, value: state.value }, props),
        activeIndex,
      };
    }
    return null;
  };

  toogleDropDown = () => {
    if (this.isMenuOpen()) {
      return this.closeMenu();
    }
    this.openMenu();
  };

  whileDropDownShow = () => {
    this.setState({
      overflow: 'visible'
    }, this.setupEventListeners);
  };

  whileDropDownHide = () => {
    this.setState({
      overflow: 'visible'
    }, this.revokeEventListeners);
  };

  afterDropDownShow = () => {
    this.setState({
      overflow: 'visible'
    });
  };

  afterDropDownHide = () => {
    this.setState({
      overflow: 'hidden'
    });
  };

  setSelectRef = select => {
    this._select = select;
    this.setState({
      menuPosition: this.getMenuPosition()
    });
  };

  setMenuRef = menu => {
    this._menu = menu;
  };

  setupEventListeners = () => {
    document.addEventListener('click', this.handleOutsideClick, false);
  };

  revokeEventListeners = () => {
    document.removeEventListener('click', this.handleOutsideClick, false);
  };

  handleOutsideClick = event => {
    if (this._menu && !this._menu.contains(event.target) && (this._input.current !== event.target) && !this._icon.contains(event.target) && this.props.closeOnOutsideClick) {
      this.closeMenu();
    }
  };

  checkMenuPosition = () => {
    const menuPosition = this.getMenuPosition();
    if (stringify(menuPosition) !== stringify(this.state.menuPosition)) {
      this.setState({
        menuPosition
      });
    }
  }

  getMenuPosition = () => {
    if (!this._select) return {};
    const height = this.props.maxDropdownHeight;
    const selectPosition = this._select.getBoundingClientRect();
    const body = document.body;
    let top = selectPosition.top + selectPosition.height;
    let dropup = false;
    if ((selectPosition.bottom + height) > (body.scrollHeight - body.scrollTop)) {
      top = top - (height + selectPosition.height);
      dropup = true;
    }
    return {
      top,
      left: selectPosition.left,
      width: selectPosition.width,
      dropup
    };
  };

  autoScroll = () => this.state.autoScroll;

  enableAutoScroll = () => {
    if (!this.autoScroll()) {
      this.setState({
        autoScroll: true
      });
    }
  };

  disableAutoScroll = () => {
    if (this.autoScroll()) {
      this.setState({
        autoScroll: false
      });
    }
  };

  handleInputChange = e => {
    e.persist();
    const value = e.target.value;
    const { onChange, onInputChange } = this.props;
    const nextState = {
      value,
      isDirty: true,
    };
    onChange([]);
    this.setState(nextState, () => {
      isMinLengthMatched(value, this.props) && this.openMenu();
      onInputChange(value, e);
    });
  }

  handleKeyDown = (e, results) => {
    let { activeIndex } = this.state;
    switch (e.keyCode) {
      case UP:
      case DOWN:
        if (!this.isMenuOpen()) {
          this.openMenu();
          return;
        }
        // Prevents input cursor from going to the beginning when pressing up.
        e.preventDefault();
        // Increment or decrement index based on user keystroke.
        activeIndex += e.keyCode === UP ? -1 : 1;
        // If we've reached the end, go back to the beginning or vice-versa.
        if (activeIndex === results.length) {
          activeIndex = -1;
        } else if (activeIndex === -2) {
          activeIndex = results.length - 1;
        }
        this.enableAutoScroll();
        this.handleActiveIndexChange(activeIndex);
        break;
      case ESC:
        this.isMenuOpen() && this.closeMenu();
        break;
      case RETURN:
        if (!this.isMenuOpen()) return;
        // Prevent form submission while menu is open.
        e.preventDefault();
        if (activeIndex !== -1 && results[activeIndex]) {
          this.handleMenuSelect(results[activeIndex]);
        }
        break;
      case TAB:
        if (!this.isMenuOpen()) {
          break;
        }
        this.closeMenu();
        break;
      default:
        break;
    }
  };

  handleActiveIndexChange = activeIndex => {
    const newState = { activeIndex };
    this.setState(newState);
  };

  handleMenuSelect = (option) => {
    const { onChange } = this.props;
    const {
      [ACTUAL_VALUE]: selected,
    } = option;
    onChange([selected]);
    this.closeMenu();
  };

  hasFocus = () => this.state.hasFocus;

  handleFocus = () => {
    if (this.hasFocus()) return;
    this.setState({
      hasFocus: true
    }, () => {
      const value = getValue(this.state, this.props);
      isSearchable(value, this.props) && this.openMenu();
      isFunction(this.props.onFocus) && this.props.onFocus(); 
    });
  };

  handleBlur = () => {
    if (!this.hasFocus()) return;
    this.setState({
      hasFocus: false
    }, () => {
      isFunction(this.props.onBlur) && this.props.onBlur();
    });
  };

  focus = () => {
    this._input.current && this._input.current.focus();
  };

  blur = () => {
    this._input.current && this._input.current.blur();
  };

  isMenuOpen = () => this.state.open;

  forceToggle = () => {
    if (!this.isMenuOpen()) {
      this.focus();
      this.openMenu();
    } else {
      this.blur();
      this.closeMenu();
    }
  }

  openMenu = () => {
    this.enableAutoScroll();
    this.checkMenuPosition();
    const {
      options,
      selected
    } = this.state;
    const activeIndex = getActiveIndex(options, selected.length && selected[0], pick(this.props, ['labelKey', 'multiple', 'defaultFirstItemSelected', 'async']));
    this.setState({
      open: true,
      activeIndex
    });
  };

  closeMenu = () => {
    this.disableAutoScroll();
    this.setState({
      open: false,
      forceOpen: false,
      activeIndex: -1
    });
  };

  render() {
    const { height, maxDropdownHeight, searchable, width, placeHolder: placeholder, loading } = this.props;
    const { overflow, hasFocus, menuPosition, results, value } = this.state;
    const menuProps = pick(this.state, OPTIONS_PROPS);
    menuProps.loading = loading;
    const iconProps = {
      className: `pk-select--icon-container${searchable ? ' has-hover-effect' : ''}`,
      ref: icon => this._icon = icon,
    };
    const inputProps = {
      ...this.props.inputProps,
      type: 'text',
      placeholder,
      className: `pk-select--input${!searchable ? ' read-only' : ''}`,
      value,
      readOnly: !searchable,
      ref: this._input,
      onFocus: this.handleFocus,
      onBlur: this.handleBlur,
      onKeyDown: e => this.handleKeyDown(e, results)
    };
    if (!searchable) {
      iconProps.onClick = () => this.forceToggle();
      inputProps.onClick = () => this.toogleDropDown();
    } else {
      iconProps.onClick = () => this.forceToggle();
      inputProps.onChange = this.handleInputChange;
    }
    return (
      <div className={`pk-select${hasFocus ? ' pk-select-focused' : ''}`} style={{ height, overflow, width }}>
        <div className='pk-select--container' ref={this.setSelectRef}>
          <div className='pk-select--value-container'>
            <div className='pk-select--input-container'>
              <input {...inputProps} />
            </div>
            {
              loading ?
                <div className='pk-select-loader-wrapper'>
                  <div className='pk-select-loader'>
                    <svg viewBox='22 22 44 44'>
                      <circle className='path' cx='44' cy='44' r='19.5' fill='none' strokeWidth='4' />
                    </svg>
                  </div>
                </div>
                :
                <div {...iconProps}>
                  <span className='pk-select-icon' style={{ backgroundImage: `url(${arrowDown})` }} />
                </div>
            }
          </div>
          <Portal container={document.body}>
            <Menu
              options={results}
              labelKey={this.props.labelKey}
              onEntering={this.whileDropDownShow}
              onExiting={this.whileDropDownHide}
              onEnter={this.afterDropDownShow}
              onExit={this.afterDropDownHide}
              maxHeight={maxDropdownHeight}
              style={menuPosition}
              menuRef={this.setMenuRef}
              selectItem={this.handleMenuSelect}
              disableAutoScroll={this.disableAutoScroll}
              {...menuProps}
            />
          </Portal>
        </div>
      </div>
    );
  }
}

export default Select;