import isEqual from 'lodash.isequal';
import isFunction from 'lodash.isfunction';
import isNumber from 'lodash.isnumber';
import isObject from 'lodash.isobject';
import isString from 'lodash.isstring';
import some from 'lodash.some';
import pick from 'lodash.pick';
import { DEFAULT_LABELKEY, ACTUAL_VALUE } from '../constants';

export const stringify = str => JSON.stringify(str);

export const addIndex = arr => arr.map((a, index) => ({ ...a, index }));

export const isSearchable = (value = '', props = {}) => props.searchable && isMinLengthMatched(value, props);

export const isMinLengthMatched = (value, props) => value.length >= props.minLength;

export const getFilterFunction = props => Array.isArray(props.filterBy) ? defaultFilterBy : props.filterBy;

export const getResults = (options, props, state) => {
  let results = [...options];
  if (!props.searchable) return results;
  if (isSearchable(state.value, props)) {
    const filterOptions = {
      ...pick(props, ['filterBy', 'labelKey', 'multiple']),
      ...pick(state, ['selected']),
      value: state.value
    };
    const filterFunction = getFilterFunction(props);
    results = options.filter(option => filterFunction(option, filterOptions));
  }
  return results;
};

export const generateItemsFromOptions = (
  options = [],
  labelKey = DEFAULT_LABELKEY,
) => options.map(option => {
  if (isString(option) || isNumber(option)) {
    return {
      [labelKey]: option,
      [ACTUAL_VALUE]: option
    };
  }
  if (isObject(option)) {
    const actual_value = {
      [ACTUAL_VALUE]: { ...option }
    };
    if (isFunction(labelKey)) {
      labelKey = labelKey(option); 
    }
    return {
      ...option,
      [labelKey]: option[labelKey],
      ...actual_value
    };
  }
});

function isMatch(input, string, props) {
  let searchStr = input;
  let str = string;

  if (!props.caseSensitive) {
    searchStr = searchStr.toLowerCase();
    str = str.toLowerCase();
  }

  return str.indexOf(searchStr) !== -1;
}

/**
 * Default algorithm for filtering results.
 */
export function defaultFilterBy(option, props) {
  const { filterBy, labelKey, multiple, selected, value } = props;

  // Don't show selected options in the menu for the multi-select case.
  if (multiple && selected.some((o) => isEqual(o, option))) {
    return false;
  }

  const fields = filterBy.slice();

  if (isFunction(labelKey) && isMatch(value, labelKey(option), props)) {
    return true;
  }

  if (isString(labelKey)) {
    // Add the `labelKey` field to the list of fields if it isn't already there.
    if (fields.indexOf(labelKey) === -1) {
      fields.unshift(labelKey);
    }
  }

  if (isString(option)) {
    return isMatch(value, option, props);
  }

  return some(fields, (field) => {
    let text = option[field];

    if (!isString(text)) {
      // Coerce to string since `toString` isn't null-safe.
      text = `${text}`;
    }

    return isMatch(value, text, props);
  });
};

export function debounce(callback, interval) {
  let debounceTimeoutId;

  return function (...args) {
    clearTimeout(debounceTimeoutId);
    debounceTimeoutId = setTimeout(() => callback.apply(this, args), interval);
  };
};
